/*
 *    X2  IRC Operator and Channel services
 *    by Rubin (rubin@afternet.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <ctype.h>
#include "config.h"
#include "sock.h"
#include "chan.h"
#include "user.h"
#include "server.h"
#include "misc.h"
#include "commands.h"
#include "parse.h"
#include "send.h"
#include "gate.h"
#include "route.h"

extern unsigned long int TotalCommands, TotalLines, TotalUserCount;
extern time_t bantimer;

int REMOTE_AUTHENTICATED = 0;  /* This gets set to TRUE when correct SERVER
                                * syntax is recieved, and we are ready to begin
                                * parsing <prefix><messages>
                                * This MUST be set back to FALSE if/when
                                * we loose our link.
                                */

/* *INDENT-OFF* */
struct
{
     char *name;
     int (*function)(M_PARAMS);
     int servercmd;
     int usercmd;
} server_commands[] =
{
/*   Command         function      servers users  .. */
  { "END_OF_BURST",  m_eob,        TRUE,     FALSE  },
  { "EOB_ACK",       m_noop,       TRUE,     FALSE  },
  { "DESYNCH",       m_desynch,    TRUE,     FALSE  },
  { "PING",          m_ping,       TRUE,     FALSE  },
  { "PONG",          m_pong,       TRUE,     TRUE   },
  { "PRIVMSG",       m_privmsg,    FALSE,    TRUE   },
  { "WC",            m_noop,       TRUE,     TRUE   },
  { "SERVER",        m_server,     TRUE,     FALSE  },
  { "BURST",         m_burst,      TRUE,     FALSE  },
  { "NOTICE",        m_ping,       TRUE,     TRUE   }, 
  { "NICK",          m_nick,       TRUE,     TRUE   },
  { "CREATE",        m_create,     FALSE,    TRUE   },
  { "JOIN",          m_join,       FALSE,    TRUE   },
  { "PART",          m_part,       FALSE,    TRUE   },
  { "MODE",          m_mode,       TRUE,     TRUE   },
  { "INVITE",        m_noop,       TRUE,     TRUE   },
  { "AWAY",          m_away,       FALSE,    TRUE   },
  { "WHOIS",         m_whois,      FALSE,    TRUE   },
  { "GLINE",         m_gline,      TRUE,     TRUE   },
  { "KICK",          m_kick,       TRUE,     TRUE   },
  { "QUIT",          m_quit,       FALSE,    TRUE   },
  { "KILL",          m_kill,       TRUE,     TRUE   },
  { "SQUIT",         m_squit,      TRUE,     TRUE   },
  { "TOPIC",         m_noop ,      TRUE,     TRUE   },
  { "NOTICE",        m_noop,       TRUE,     TRUE   },
  { "INFO",          m_noop,       TRUE,     TRUE   },
  { "ADMIN",         m_noop,       TRUE,     TRUE   },
  { "CONNECT",       m_noop,       TRUE,     TRUE   },
  { "TIME",          m_noop,       TRUE,     TRUE   },
  { "TRACE",         m_noop,       TRUE,     TRUE   },
  { "MOTD",          m_noop,       TRUE,     TRUE   },
  { "RPING",         m_noop,       TRUE,     TRUE   },
  { "SILENCE",       m_noop,       TRUE,     TRUE   },
  { "UPING",         m_noop,       TRUE,     TRUE   },
  { "VERSION",       m_noop,       TRUE,     TRUE   },
  { "WALLOPS",       m_noop,       TRUE,     TRUE   },
  { "STATS",         m_noop,       TRUE,     TRUE   },
  { "SETTIME",       m_settime,    TRUE,     FALSE  },
  { NULL,            NULL,         FALSE,    FALSE  }
};
/* *INDENT-ON* */ 

/**************************************************
 * Beginning m_function definitions. These can be *
 * anything that is recieved as a command from    *
 * the server.                                    *
 * All the m_functions use the same               *
 * prototype M_PRAMS which looks like this:       *
 ******************************************************************
 * int m_xxxx(ServerListType sptr, UserListType uptr, char* rest);*
 ******************************************************************/

/* A dummy command for stuff we understand
 * but dont need to parse
 */
int m_noop(M_PARAMS)
{
  return(0);
}
int m_eob(M_PARAMS)
{
   /* Only send ack to our uplink. */
   if(!strcasecmp(sptr->Name, Uplink))
   {
     JoinChans();
     send_eob_ack();
   }
   else
   {
     Debug(DBGINFO, "NOT responding to EOB from non-uplink server");
   }
   return(0);
}
int m_desynch(M_PARAMS)
{
  Debug(DBGINFO, "<%s> DESYNCH %s", sptr->Name, rest);
  return(0);
}
int m_ping(M_PARAMS)
{
   if(rest++)
      send_serverpong(rest);
   return(0);
}
int m_privmsg(M_PARAMS)
{
   char* to;
   to = (char *) strtok(rest, " ");
   rest = (char *) strtok(NULL, "\r\n\0");
   if(rest[0] == ':')
     rest++;
   if(to[0] == '#')
   {
     if(rest[0] == CommandMark)
     {
        ProcessCommand(uptr, to, rest);
     }
   }
   else if(to[0] == convert2y[ServiceNumeric])
   {
      Debug(DBGINFO, "m_privmsg() %s(%s) sent privmsg to %s [%s]", uptr->Nick, uptr->Numnick, to, rest); 
     /* Check for CTCP */
     if(rest[0] == CTCP)
     {
        if(rest[strlen(rest)-1] != CTCP)
        {  /* No terminator on CTCP, ignore it */
          Debug(DBGINFO, "Ignoring unterminated CTCP");
          return(0);
        }
        rest++;
        rest[strlen(rest)-1] = '\0';
        ProcessCTCP(uptr, to, rest);
        return(0);
     }
     else
     {
        ProcessCommand(uptr, to, rest);
        return(0);
     }
   }
   return(0);
}
int m_server(M_PARAMS)
{
    char* Arg[8];
    Arg[0] = (char *) strtok(rest, " ");  /* Name */
    Arg[1] = (char *) strtok(NULL, " ");  /* hopes */
    Arg[2] = (char *) strtok(NULL, " ");  /* ts 1 */
    Arg[3] = (char *) strtok(NULL, " ");  /* ts 2 */
    Arg[4] = (char *) strtok(NULL, " ");  /* Protocol (P10) */
    Arg[5] = (char *) strtok(NULL, " ");  /* <numnick><masknum> */
    Arg[6] = (char *) strtok(NULL, " ");  /* 0 */
    Arg[7] = (char *) strtok(NULL, "\0\r\n"); /* Description */

    if(!(Arg[6] && Arg[7]))
    {
        Debug(DBGWARNING, "Incorrect syntax in SERVER line");
        return(0);
    }
    AddServer(Arg[0], sptr->Name, Arg[5][0]);
    return (0);
}
/* E.G.:
 * BURST #AfterNET 959324000 +tnl 23 kAC:o,kAB,CAA,kAA :%*!*fo4o@*foo.com *!*fo3o@*foo.com *!*fo2o@*foo.com *!*foo@*foo.com
 */
int m_burst(M_PARAMS)
{
    char* channel;
    char* timestamp;
    int   ts = 0;
    char* parm;
    ChannelListType ChanPtr;

    channel = strtok(rest, " ");
    timestamp = strtok(NULL, " ");
    if(!(*timestamp))
    {
       Debug(DBGWARNING, "Problem parsing burst. No 2nd parameter.");
       return(0);
    }
    if((ChanPtr = GetChannel(channel)))
    {
      if((ts = atoi(timestamp)))
      {
        if(!strcasecmp(Uplink, sptr->Name))
        {
          /* If the uplink is sending it */
          ChanPtr->CreationDate = ts;
        }
        else
        {
           /* Some other server somewhere is connecting */
           if(ChanPtr->Regged != 0 && ChanPtr->settings[INCHAN] == 1)
           {
             if(ts < ChanPtr->CreationDate)
                 send_opchan(ChanPtr);
           }
        }
      }
      else
      {
        Debug(DBGWARNING, "Invalid TS during BURST.. duno what to do!");
        return(0);
      }
    }
    else  /* Its not a channel I have in memory.. */
    {
        if (!AddChan(channel, FALSE, 0, NULL, NULL, NULL, NULL, NULL,NULL, NetTime(), 0))
        {
           Debug(DBGWARNING, "Error adding channel %s in m_burst()", channel);
           return(0);
        }
        ChanPtr = GetChannel(channel);
    }
    for(parm = strtok(NULL, " "); /* parm = first param */
        parm;   /* While param is not null */
        parm = strtok(NULL, " ")  /* parm = next param */
       )
    {
       if(parm[0]==':')  parm++;
       switch(parm[0])
       {
          case '+': /* Additive channel modes. l and k both cause an additional paramiter */
          {
              int i; char* key = NULL; char* limit = NULL;
              char buff[MAXLEN];
              for(i=0;parm[i];i++)
              {
                  if(parm[i] == 'k')
                       key = strtok(NULL, " ");
                  if(parm[i] == 'l')
                       limit = strtok(NULL, " ");
             }
              /* Now we have eg "MODE #channel parm key limit" */
              /* p10.html seems to say key is always first     */
              sprintf(buff, "%s %s %s", parm, key?key:"", limit?limit:"");
              on_mode(ChanPtr, uptr, buff);
              break;
          }
          case '%': /* burst ban list */
          {
//              ChannelListType ChanPtr;
              char* ban;
//              if(!(ChanPtr = GetChannel(channel)))
//                  return(0);  /* Not a channel we are in */
              parm++; /* Ignore % */
              for( ban = parm;      /* ban = first ban */
                   ban;             /* while ban != null */
                   ban = (char *) strtok(NULL, " ")  /* Get next ban from outside stack */
                 )
              {
                   add_ban(ChanPtr, ban);
              }
              break;
          }
          default:  /* Comma seperated list of nicks and there modes */
          {
              char* mode=""; char* numnick;
              /* First off, we have a comma seperated list.. */
              for( numnick = (char *) mysep(&parm, ",");  /* Nick = first NUM nick */
                   numnick;                               /* While theres a NUM nick.. */
                   numnick = (char *) mysep(&parm, ",")  /* Nick = next NUM nick */
                 )
              {
                 if((mode = strchr(numnick, ':')))
                 {
                       *mode = '\0';
                       mode++;
                 }
                 ProcessJoin(ChanPtr, FindNumnick(numnick, UsersU), ChanPtr->Name, 'B');
                /* Should we on_mode here as well? */ 
                /* My thinking is not, because who wants x2
                 * showing up after a split and changing things.
                 */
              }
              break;
          }
              
       }
    }
    return(0);
}
int m_nick(M_PARAMS)
{
    char* Arg[9];
    char* modes;
    UserListType UserPtr;
    unsigned int IP;
    Arg[0] = (char *) strtok(rest, " ");   /* nick */
    Arg[1] = (char *) strtok(NULL, " ");   /* 1 */
    Arg[2] = (char *) strtok(NULL, " ");   /* ts */
    if (Arg[2] == NULL)
    {
        if(uptr)
        {
          ProcessNickChange(uptr->Nick, Arg[0]);
        }
        else
        {
           Debug(DBGWARNING, "NICK: Too few arguments!");
        }
        return (0);
    }
    Arg[3] = (char *) strtok(NULL, " ");   /* userID */
    Arg[4] = (char *) strtok(NULL, " ");   /* address */
    modes = (char *) strtok(NULL, " ");   /* modes */
    if(modes[0] == '+')
    {
       Arg[5] = modes;
       Arg[6] = (char *) strtok(NULL, " "); /* Compressed IP */
    }
    else
    {
       Arg[5] = "";
       Arg[6] = modes;
    }
    /* Need to convert Arg[6] to IP here.. */
    IP = base64toint(Arg[6]);
    /*************/
    Arg[7] = (char *) strtok(NULL, " ");  /* numnick */
    Arg[8] = (char *) strtok(NULL, "\n");  /* :RealName */
    if(Arg[8] && *Arg[8])
       Arg[8]++;
    TotalUserCount++;
    if(!(Arg[7]))
    {
        Debug(DBGWARNING, "Incorrect syntax in NICK line");
        return(0);
    }
    UserPtr = AddUser(Arg[7], Arg[0], Arg[3], Arg[4], Arg[8], IP, sptr->Name, NOMODES);
    if(!UserPtr)
    {
      Debug(DBGWARNING, "AddUser returned NULL for %s", Arg[0]);
      return(0);
    }
    if(modes[0] == '+')
      ProcessUserMode(UserPtr, modes);
    if(!NETBURSTING)
      SendMOTD(UserPtr);
      if(UserPtr)
      {
        char* dq = unsigned32ToDotQuad(IP);
        if(!GlineCheck(UserPtr, dq) && !SuperGlineCheck(UserPtr, dq))
          BeginGateScan(UserPtr, dq);
      }
    return (0);
}
/* CREATE */
int m_create(M_PARAMS)
{
   int i;
   char *channel;

   channel = (char *) strtok(rest, " ");
   if(!channel)
   {
        Debug(DBGWARNING, "Invalid JOIN syntax (null channel).");
        return(0);
   }
   /* Need to check for comma seperated channels */
   for((i=strlen(channel)-2);(i >= 0);i--)
   {/* move right to left.. call ProcessJoin when we hit a comma or beginning */
        if(channel[i] == ',')
        {
          ChannelListType cptr;
          cptr = GetChannel(&channel[i+1]);
          ProcessJoin(cptr, uptr, &channel[i+1], 'C');
          channel[i] = '\0';
        }
   }
   ProcessJoin(GetChannel(channel), uptr, channel, 'C');
   return(0);
}
/* JOIN */
int m_join(M_PARAMS)
{
   int i;
   char *channel;

   channel = (char *) strtok(rest, " ");
   if(!channel)
   {
        Debug(DBGWARNING, "Invalid JOIN syntax (null channel).");
        return(0);
   }
   /* Need to check for comma seperated channels */
   for(i=strlen(channel)-2;i>=0;i--)
   {/* move right to left.. call ProcessJoin when we hit a comma or beginning */
        if(channel[i] == ',')
        {
          ChannelListType cptr;
          cptr = GetChannel(&channel[i+1]);
          if(!strcmp(&channel[i+1], "0"))
          {
             RemoveNickFromAllChannels(uptr);
          }
          else
          {
            ProcessJoin(cptr, uptr, &channel[i+1], 'J');
          }
          channel[i] = '\0';
        }
   }
   if(!strcmp(channel, "0"))
   {
      /* JOIN 0 means part all join'd channels. */
      RemoveNickFromAllChannels(uptr);
   }
   else
   {
      ProcessJoin(GetChannel(channel), uptr, channel, 'J');
   }
   return(0);
}
/* PART */
int m_part(M_PARAMS)
{
     int i;
     ChannelListType cptr;
     char* channel;
     channel = (char *) strtok(rest, " ");
     if(!channel)
     {
        Debug(DBGWARNING, "Invalid PART syntax.");
        return(0);
     }
     /* Need to check for comma seperated channels */
     for(i=strlen(channel)-2;i>=0;i--)
     {/* move right to left.. call ProcessPart when we hit a comma or beginning */
        if(channel[i] == ',')
        {
             if((cptr = GetChannel(&channel[i+1])))
             {
               ProcessPart(cptr, uptr);
             }
             else
             {
                Debug(DBGWARNING, "m_part(): cant find [%s] in memory.", &channel[i+1]);
             }
             channel[i] = '\0';
        }
     }
     if((cptr = GetChannel(channel)))
     {
        ProcessPart(cptr, uptr);
     }
     else
     {
        Debug(DBGWARNING, "m_part(): can't find [%s] in memory.", channel);
     }
     return(0);
}
/* MODE */
int m_mode(M_PARAMS)
{
         char* Arg[4];

         Arg[0] = (char *) strtok(rest, " ");   /* nick/channel */
         Arg[1] = (char *) strtok(NULL, " ");   /* +/- mode   */
         if (Arg[0][0] == '#')
         {
             char buff[MAXLEN];
             Arg[2] = (char *) strtok(NULL, " ");
             Arg[3] = (char *) strtok(NULL, "\n\0");
             snprintf(buff, MAXLEN, "%s %s %s", Arg[1], Arg[2]?Arg[2]:"", Arg[3]?Arg[3]:"");
                        /* "channel +s+nkl stuff stuff" */
             if(uptr)
               on_mode(GetChannel(Arg[0]), uptr, buff);
             else
               on_mode(GetChannel(Arg[0]), NULL, buff);
         }
         else
             ProcessUserMode(uptr, &Arg[1][1]);
         return (0);
}
int m_pong(M_PARAMS)
{
    GotPong = true;
    PongStamp = time(NULL);  /* does it matter who sent us the pong? */
    return (0);
}
int m_away(M_PARAMS)
{
     char* reason;
     reason = (char *)strtok(rest, " ");  /* :Reason */
     if(reason)
       SetAway(uptr);
     else
       ClearAway(uptr);
     return(0);
}
/* SETTIME */
int m_settime(M_PARAMS)
{
     char *str;
     str = (char *) strtok(rest, " ");
     if(str)
     {
        NetOffset = atoi(str) - time(NULL);
     }
     return(0);
}
/* WHOIS */
int m_whois(M_PARAMS)
{
    send_numeric(311, uptr->Numnick, "%s %s %s * :\002%c\002) /MSG %s HELP (\002%c\002",
       IrcNick, IrcNick, ServiceName, CommandMark, IrcNick, CommandMark);
    send_numeric(312, uptr->Numnick, "%s %s :%s",
       IrcNick, ServiceName, VERSION);
    send_numeric(313, uptr->Numnick, "%s :is an AfterNet IRC Operator",
       IrcNick);
    send_numeric(317, uptr->Numnick, "%s 0 %d :seconds idle, signon time",
       IrcNick, ConnectTime);
    send_numeric(318, uptr->Numnick, "%s :End of /WHOIS list.",
       IrcNick);
    return (0);
}
/* GLINE */
int m_gline(M_PARAMS)
{
    char* Arg[6];
    Arg[2] = (char *) strtok(rest, " ");        /* Target (*) */
    Arg[3] = (char *) strtok(NULL, " ");        /* +/-*@* */
    Arg[4] = (char *) strtok(NULL, " ");        /* Time */
    Arg[5] = (char *) strtok(NULL, "\0");       /* :<nick> Reason */
    if (Arg[3][0] == '+')
    {
      static time_t gwritets=0;
      if (!Arg[4])
    	return (0);		/* Invalid no time.. */
      ProcessAddGline(&Arg[3][1], atoi(Arg[4]), Arg[5] + 1);
      if(time(NULL) > gwritets + 60)
      {
         WriteGlineDat();
         gwritets=time(NULL);
      }
    }
    else if (Arg[3][0] == '-')
      ProcessRemoveGline(&Arg[3][1]);
    return (0);
}
int m_kick(M_PARAMS)
{
     char* Arg[4];
     ChannelListType cptr;
     Arg[2] = (char *) strtok(rest, " "); /* Channel */
     Arg[3] = (char *) strtok(NULL, " "); /* Target */
     if(!Arg[3])
     {
        Debug(DBGWARNING, "Incorrect KICK syntax in line.");
        return(0);
     }
     cptr = GetChannel(Arg[2]);
     if(cptr == NULL)
        return(0);
     ProcessKick(cptr, uptr, Arg[3]);
     return (0);
}
int m_quit(M_PARAMS)
{
     DeleteUser(uptr, true);
     return (0);
}
int m_kill(M_PARAMS)
{
     char* targetnick;
     UserListType target;
     if((targetnick = (char *) strtok(rest, " ")))
     {
        if(!(target = FindNumnick(targetnick, UsersU)))
        {
           Debug(DBGWARNING, "m_kill(): target %s not found", targetnick);
           return(0);
        }
        DeleteUser(target, true);
     }
     else
        Debug(DBGWARNING, "Incorrect syntax for KILL in line");
     return(0);
}
int m_squit(M_PARAMS)
{
     ServerListType ServerPtr;
     char* Arg[5];
     RouteListType rptr;

     Arg[2] = (char *) strtok(rest, " "); /* Server name */
     Arg[3] = (char *) strtok(NULL, " "); /* Time stamp */
     Arg[4] = (char *) strtok(NULL, "\n"); /* Reason */  
     if (!(Arg[2]))
     {
         Debug(DBGWARNING, "Incorrect syntax for SQUIT in line");
         return (0); 
     }
     if (!strcasecmp(Arg[2], Uplink))
     {
           /* If its us, handle it.. */
           Debug(DBGWARNING, "We were SQUIT, killing link..");
           SockConnected = FALSE;
           return(0);
     }
    /* Profiling - adjust karma */
    ServerPtr = GetServer(Arg[2]);
    if(!ServerPtr)
    {
       Debug(DBGWARNING, "Can't find server %s to squit!", Arg[2]);
       return(1);
    }
    rptr = FindRouteListServer(Route, ServerPtr->Name);
    if(!strcasecmp(Arg[4], ":Ping Timeout"))
    {
        /* Adjust karma down */
        if(rptr)
           rptr->karma--;
        if(PINGOUTCONNECT)
        {
              if(rptr) /* If we have a listing for this one */
              {
                 RouteListType rptr_1, rptr_2, rptr_connect;

                 rptr_1 = FindRouteListServer(Route, rptr->uplink);
                 rptr_2 = FindRouteListServer(Route, rptr->secondaryuplink);
                 if(rptr_1 && GetServer(rptr_1->server))
                 {
                   if(rptr_2 && GetServer(rptr_2->server))
                   {
                     /* +2 to give primary slight advantage */
                     if(rptr_1->karma+2 >= rptr_2->karma)
                         rptr_connect = rptr_1;
                     else
                         rptr_connect = rptr_2;
                   }
                   else
                     rptr_connect = rptr_1;
                   send_connect(ServerPtr->Name, rptr->port, GetServer(rptr_connect->server));
                 }
                 else
                    send_connect(ServerPtr->Name, rptr->port, GetServer(ServerPtr->LinkName));
              }
              else
              {
                 /* Fall back on connecting it to where it was. */
                 send_connect(ServerPtr->Name, 8888, GetServer(ServerPtr->LinkName));
              }
        }
    }  
    RemoveServer(Arg[2]);
    return (0);
}

/**************************************************
 * Beginning general parse functions which        *
 * eventually get process flow to the             *
 * above m_functions                              *
 **************************************************/

int ProcessNumeric(ServerListType sptr, int num, UserListType target, char* rest)
{
  Debug(DBGINFO, "Processing Numeric: <%s> %d <%s> %s", sptr->Name, num, target?target->Nick:"", rest);
  switch (num)
  {
    case 1:			/* welcome */
    case 2:			/* Your host.. */
    case 3:			/* Created... */
    case 4:			/* MYINFO */
    case 5:			/* Map reply */
    case 6:			/* Map More */
    case 7:			/* Map End */
    case 8:			/* Server Notice Mask */
    case 9:			/* StatMem Total */
    case 10:			/* StatMem */

    case 200:			/* Trace Link */
    case 201:			/* Trace Connecting.. */
    case 202:			/* Trace Handshake */
    case 203:			/* Trace Unknown */
    case 204:			/* Trace oper */
    case 205:			/* Trace User */
    case 206:			/* Trace Server */
    case 208:			/* TraceNewType */
    case 209:			/* Trace Class */
    case 211:			/* Stats Link Info */
    case 212:			/* Stats commands */
    case 213:			/* Stats C line */
    case 214:			/* Stats N line */
    case 215:			/* Stats I line */
    case 216:			/* Stats K line */
    case 217:			/* Stats P line */
    case 218:			/* Stats Y line */
    case 219:			/* end of Stats */
    case 220:			/* NULL */
    case 221:			/* UMODEIS */
    case 231:			/* Service Info */
    case 232:			/* End of Service Info */
    case 233:			/* Service */
    case 234:			/* Serv list */
    case 235:			/* Serv list End */
    case 241:			/* Stats L line */
    case 242:			/* Stats Uptime */
    case 243:			/* Stats O line */
    case 244:			/* Stats H line */
    case 246:			/* Stats T lien */
    case 247:			/* Stats G line */
    case 248:			/* Stats U line */
    case 250:			/* Stats Conn */
    case 251:			/* LuserClient */
    case 252:			/* Luser Op */
    case 253:			/* Luser Unknown */
    case 254:			/* Luser Channels */
    case 255:			/* Luser Me */
    case 256:			/* Admin */
    case 257:			/* AdminLOC1 */
    case 258:			/* AdminLOC2 */
    case 259:			/* Admin Email */
    case 261:			/* TraceLog */
    case 262:			/* TracePing */
    case 271:			/* Silent list */
    case 272:			/* End of silent list */
    case 275:			/* Stats D lien */
    case 280:			/* Glist */
    case 281:			/* End of G list */
    case 283:			/* QUAR Nick */
      return (0);

    case 301:			/* away reply */
    {
         /* Ok, so now we send "WHOIS nick nick" when we see people oper */
         /* Then here we see them away .. so after a net join .IRCOPS */
         /* knows who'se here and who'se not */
      char* who;
      who = (char *) strtok(rest, " ");
      SetAway(FindNick(who, UsersN));
      return(0);
    }
    case 302:			/* UserHost reply */
    case 303:			/* ISon reply */
    case 304:			/* RPL_TEXT */
    case 305:			/* No longer away */
    case 306:			/* Now marked as away */
    case 307:			/* userip reply */
    case 311:			/* WHOIS USER */
    case 312:			/* whois server */
    case 313:			/* whois oper */
    case 314:			/* whowas user */
    case 315:			/* End of WHO */
    case 316:			/* NULL */
    case 317:			/* whois Idle */
    case 318:			/* End of whois */
    case 319:			/* whois chans */
    case 320:			/* NULL */
    case 321:			/* List Start */
    case 322:			/* RPL_LIST */
    case 323:			/* List end */
    case 324:			/* Channel Mode Is... */
    case 325:                   /* Unknown */
    case 326:                   /* Unknown */
    case 327:                   /* Unknown */
    case 328:                   /* DalNet Channel URL */
    case 329:			/* Creation Time */
    case 331:			/* No Topic */
    case 332:			/* RPL_TOPIC */
    case 333:			/* Topic who time */
    case 334:			/* List Usage */
    case 341:			/* RPL_Inviting */
    case 351:			/* Version reply */
    case 352:			/* Who reply */
    case 353:			/* Name Reply */
    case 354:			/* Whospcrpl */
    case 361:			/* KILL DONE */
    case 362:			/* CLOSING */
    case 363:			/* CloseEnd */
    case 364:			/* LINKS */
    case 365:			/* End of links */
    case 366:			/* End of Names */
    case 367:			/* Banlist */
    case 368:			/* End of Banlist */
    case 369:			/* End of WhoWas */
    case 371:			/* INFO */
    case 372:			/* MOTD */
    case 373:			/* InfoStart */
    case 374:			/* End of Info */
    case 375:			/* MOTD Start */
    case 376:			/* End of MOTD */
    case 381:			/* Your nowan ircop */
    case 382:			/* Rehashing */
    case 384:			/* My Port Is */
    case 385:			/* Not Oper anymore */
    case 391:			/* Time reply */
      return (0);
    case 401:			/* No Such Nick */
    case 402:			/* No such Server */
    { /* Often this was from us trying to connect a server (lack of c/n on connecting side) */
         char* nick;
         char* commandstr;
         char* errmsg;
         char* server;

         Debug(DBGINFO, "402: %s says: %s", sptr->Name, rest);
         nick = (char *) strtok(rest, " ");
         commandstr = (char *) strtok(NULL, " ");
         errmsg = (char *) strtok(NULL, "\0");
         if(sptr && !strcasecmp(commandstr, ":CONNECT:"))
         {
           server = (char *) strtok(errmsg, " "); /* the word "Host" */
           server = (char *) strtok(NULL, " ");  /* hostname who'se missing */
           if(!server || !*server)
           {
               Debug(DBGWARNING, "empty target server");
               return(0);
           }
           /* TODO : somehow mark that that uplink doesnt work for this server.
            *        LOG it, and complain.  Someone is missing c/n.
            */
           Debug(DBGWARNING, "Uplink missing C/N lines for %s!", server);
           HandleFailedConnect(sptr, server);
         }
         else if(sptr && !strcasecmp(commandstr, ":link"))
         {
             /* :Link with mephisto.etheria.cx cancelled: Server mephisto.etheria.cx[216.46.33.71]
              * :Link with mephisto.etheria.cx established. :No such server
              */
               server = (char *) strtok(errmsg, " ");  /* with            */
               server = (char *) strtok(NULL, " ");    /* servername      */
               errmsg = (char *) strtok(NULL, " ");    /* cancelled / established */
               if(!server)
               {
                   Debug(DBGWARNING, "402 :link...: empty target server");
                   return(0);
               }
               if(!strcasecmp(errmsg, "cancelled:"))
               {
                  errmsg = (char *) strtok(NULL, ":");
                  Debug(DBGINFO, "Failed to connect %s: %s", server, errmsg);
                  HandleFailedConnect(sptr, server);
               }
         }
         return(0);
    }
    case 403:			/* No such channel */
      return (0);
    case 404:
    {
         char* Channel;
         Channel = (char *) strtok(rest, " ");
         Debug(DBGWARNING, "Error 404: Cannot send to channel %s", Channel);
         return (0);
    }
    case 405:			/* Too many Channels */
    case 406:			/* Was no such nick */
    case 407:			/* TooManyTargets */
    case 408:			/* NULL */
    case 409:			/* No Origin */
    case 410:			/* NULL */
    case 411:			/* NO Recipient */
    case 412:			/* No text to send */
    case 413:			/* No top level domain */
    case 414:			/* Wildcard in toplevel domain */
    case 415:			/* NULL */
    case 416:			/* Query too long */
    case 417:			/* NULL */
    case 421:			/* Unknown command */
    case 422:			/* No MOTD */
    case 423:			/* No Admin Info */
    case 424:			/* NULL */
    case 431:			/* No Nick Given */
    case 432:			/* Erroneus nick name */
    case 433:			/* Nick name in use */
    case 434:			/* NULL */
    case 435:			/* NULL */
    case 436:			/* Nick Collision */
    case 437:			/* Baned user nick change error */
    case 438:			/* Nick change to fast */
    case 439:			/* Target changed to fast */
    case 440:			/* NULL */
    case 441:			/* User Not In Channel */
    case 442:			/* You arnt on channel */
    case 443:			/* Already in channel */
    case 444:			/* NULL */
    case 451:			/* Not registered */
    case 452:			/* NULL */
    case 461:			/* Need more params */
    case 462:			/* Already registered */
    case 463:			/* No permmision for host */
    case 464:			/* Pass mismatch */
    case 465:			/* Your banned creep */
    case 466:			/* You will be banned */
    case 467:			/* Chan key already set */
    case 468:			/* Invalid username */
    case 469:			/* NULL */
    case 470:			/* NULL */
    case 471:			/* Chan full */
    case 472:			/* Unknown mode */
    case 473:			/* Invite only chan */
    case 474:			/* Banned from chan */
    case 475:			/* Bad key */
    case 476:			/* Bad Channel Mask */
    case 477:			/* Chan does not support modes */
    case 478:			/* Banlist full */
    case 479:			/* NULL */
    case 480:			/* NULL */
    case 481:			/* No Privlidges (not oper) */
    case 482:			/* Your not channel op */
    case 483:			/* Cant kill a server stupid */
    case 484:			/* cant kill/kick/deop chan service (+k) */
    case 485:			/* NULL */
    case 491:			/* NO Olines for your host */
    case 501:			/* Unknown Mode flag */
    case 502:			/* Cant change other users modes */
    case 511:			/* Silent list full */
    case 512:			/* No such gline */
    case 513:			/* Bad Ping */
    case 514:			/* Cant kill ircop (afternet) */
      return (0);
    default:
      return (0);
  }
}

int ProcessCommand(UserListType from, char *to, char *rest)
{
   ChannelListType ChanPtr = NULL;
   char *Chan = "";
   char *Command;
   int public = 0;

   if(to[0] == '#')
   { /* A public command to a channel */
      if(rest[0] == CommandMark)
      {
         rest++;
         ChanPtr = GetChannel(to);     
         if(!ChanPtr)
         {
            Debug(DBGWARNING, "Recieved channel chatter when i dont have that channel in my DB");
            return(0);
         }
      }
      else
        return(0);  /* Msg to channel. ignore.. */
      public = TRUE;
   }
   if(rest && rest[0] == '#')
   {
      Chan = (char *) strtok(rest, " ");
      rest = (char *) strtok(NULL, "\r\n\0");
      ChanPtr = GetChannel(Chan);
   }
   else if(public)
   {
      Chan = to;
   }
   Command = (char *) strtok(rest, " ");
   if(!Command)
   {
      Debug(DBGINFO, "Empty command. ignoring.");
      return(0);
   }
   rest = (char *) strtok(NULL, "\0\r\n");
/* OK well make the old way work too.. */
   if(rest && rest[0] == '#')
   {
      Chan = (char *) strtok(rest, " ");
      rest = (char *) strtok(NULL, "\r\n\0");

      ChanPtr = GetChannel(Chan);
   }
   on_msg(from, Chan, to, public, Command, rest);
   return(0);
}

int ProcessCTCP(UserListType from, char *to, char *rest)
{
    char *command;

    command = (char *) strtok(rest, " ");
    rest = (char *) strtok(NULL, "\0\r\n");
    if(!command)
    {
       return(0); /* empty ctcp */
    }
    if(!strcasecmp(command, "VERSION"))
    {
       send_ctcp_version_reply(from);
       return(0);
    }
    else if(!strcasecmp(command, "PING"))
    {
       if(!rest)
         return(0); /* empty ping request */
       send_ctcp_ping_reply(from, rest);
       return(0);
    }
    else if(!strcasecmp(command, "CLIENTINFO"))
    {
      send_ctcp_clientinfo_reply(from);
      return(0);
    }
    else if(!strcasecmp(command, "TIME"))
    {
      send_ctcp_time_reply(from);
      return(0);
    }
   return(0);
}

int ProcessMsg(UserListType from, char *to, char *rest)
{
   /* Was it done in a channel or to me personaly? */
   /* Just a thought.. is it possible to get NICK PRIVMSG #channel1,#channel2 etc? */
   if(rest[0] == ':')
      rest++;
   /* Check for CTCP */
   if(rest[0] == CTCP)
   {
      if(rest[strlen(rest)-1] != CTCP)
      {  /* No terminator on CTCP, ignore it */
        Debug(DBGINFO, "Ignoring unterminated CTCP");
        return(0);
      }
      rest++;
      rest[strlen(rest)-1] = '\0';
      ProcessCTCP(from, to, rest);
      return(0);
   }
   else
   {
      ProcessCommand(from, to, rest);
      return(0);
   }
}

void ProcessPreAuthLine(char *LineBuff)
{
   char *Arg[MAXLEN];
   static char RemotePass[MAXPASS]; /* should be this or just '10'? */

   Arg[0] = (char *) strtok(LineBuff, " ");
   Arg[1] = (char *) strtok(NULL, " ");

   if(!strcmp(Arg[0], "PING"))
   {    /* Respond with a ping reply */
        send_preauth_pong(Arg[1]);
        return;
   }
   else if (!strcmp(Arg[0], "ERROR"))
   {    /* Error handling should be implimented here.. */
	return;  /* need to figureo ut which errors to ignore */
        if(!(Arg[2] = (char *) strtok(NULL, "\0")))
           Arg[2] = "";
        Debug(DBGSEVERE, "ERROR from socket! :%s %s %s", Arg[0], Arg[1], Arg[2]);
        SockConnected = FALSE;
        return;
   }
   else if (!strcmp(Arg[0], "PASS"))
   {    /* store the password for compareing later when we get SERVER */
         Debug(DBGNOTICE, "Got PASS from server: %s", Arg[1]);
         safecopy(RemotePass, ++Arg[1], 10);
         return;
   }
   else if (!strcmp(Arg[0], "SERVER"))
   {    /* Beginning of burst. validate password and name, authenticate and send our burst
            * if we were listning for the link */
         UpLinkList *UpLinkPtr;
	 char Numnick[NUMNICKLEN];
    
         Debug(DBGNOTICE, "Got SERVER line from uplink. Processing.");
         Arg[2] = (char *) strtok(NULL, " ");         /* Hops */
         Arg[3] = (char *) strtok(NULL, " ");         /* TS #1 */
         Arg[4] = (char *) strtok(NULL, " ");         /* TS #2 */
         Arg[5] = (char *) strtok(NULL, " ");         /* Protocol */
         Arg[6] = (char *) strtok(NULL, " ");         /* CD] masknick */
         Arg[7] = (char *) strtok(NULL, "\r\n\0");      /* Description */
         if(!(Arg[7] && Arg[6]))        
         {
            Debug(DBGWARNING, "Invalid SERVER recieved.  Killing link.");
            SockConnected = FALSE;
            return;
         }
         Debug(DBGINFO, "PROTOCOL: %s", Arg[5]);
         if(strcmp(Arg[5], "J10")) /* If Arg[5] is not "J10" then.. */
         {
             Debug(DBGWARNING, "UNKNOWN protocol. This is NOT a compatable irc daemon.");
             Debug(DBGWARNING, "Currently X2 works with the following irc servers:");
             Debug(DBGWARNING, "   Universal ircd  - http://ircu.sourceforge.net");
             Debug(DBGWARNING, "OLD Undernet ircd  - 2.10.07 and lower P10 versions.");
             Debug(DBGWARNING, "");
             Debug(DBGWARNING, "Making x2 work with other versions requires serious rewriting.");
             sleep(4);
         }
         for(UpLinkPtr = UpLinks;UpLinkPtr;UpLinkPtr = UpLinkPtr->Next)
         {
             Debug(DBGINFO, "Comparing incomming connection (%s) to our (%s)", Arg[1], UpLinkPtr->ServName);
             if(!strcasecmp(UpLinkPtr->ServName, Arg[1]))  /* If they match */
             {
                 /* We have a match, check password */
                 if(strcmp(RemotePass, UpLinkPtr->ServPass))
                 {  /* If the passwords dont match.. */
                    Debug(DBGWARNING, "Password for server %s doesnt match. Killing link.", Arg[1]);
                    send_squit(Arg[1], "Password mismatch");
                    SockConnected = FALSE;
#ifdef debugon
                    exit(2);
#endif
                    return;
                 }
                 Debug(DBGINFO, "Adding server %s %s %c", Arg[1], ServiceName, Arg[6][0]);
                 AddServer(Arg[1], ServiceName, Arg[6][0]);
                 NETBURSTING = TRUE;  /* Ignore incomming users for a while*/
                 REMOTE_AUTHENTICATED = TRUE;  /* Stop parsing PASS/SERVER etc */
                 bantimer = time(NULL);  /* Start ban timer over */
                 NetOffset = atoi(Arg[4]) - time(NULL);  /* Get the network timestamp from the server */
                 ServerConnected = TRUE;  /* We now have a good link.. */
                 safecopy(Uplink, Arg[1], MAXSERVERLEN);
                 if(!SentBurst) /* If we were listning, we must now return the burst.. */
                 {
                    if(!Authenticated)
                    {
                       Authenticate(UpLinkPtr);
                       ChangeRouteUplinks(Route);
                       RankOutsideNess(Route);

                    }
                 }
                 strcpy(Numnick, "?X2");
                 Numnick[0] = convert2y[ServiceNumeric]; /* Server numnick is first letter of user numnick */
                 AddUser(Numnick, IrcNick, IrcNick, "X2", ServiceName,
                         LOCALIP, ServiceName, (UMODEk & UMODEo));
                 ProcessUserMode(FindNick(IrcNick, UsersN), "+kows");

                 return;
             }
        }
        Debug(DBGWARNING, "Warning server %s is not in our list. Killing link.", Arg[1]);
        send_squit(Arg[1], "No C/N conf lines");
        SockConnected = FALSE;
        sleep(4); /* Lets not flood the server with connect requests TOO fast ;) */
        return; /* Servername not in our list */
   }
   else if (!strcasecmp(Arg[0], "SQUIT"))
   {   /* Server quit: Clear all the users and server of this
           * and everyone behind it. */
          /* I think SQUIT's generally happy as Arg[1] so this shouldnt be used.. */
        RemoveServer(Arg[1]);
        return;
   }
}

  /* the <prefix> here could still be
   * :<nick> or <numnick>.  so we'll
   * check that, look up the senders
   * record (be they nick or server)
   * and then go about parsing stuff.
   */
int ProcessMessage(char *LineBuff)
{
  char* Sender;
  char* Command;
  char* rest;
  UserListType UserPtr = NULL;
  ServerListType ServerPtr = NULL;
  int i;
  /* prefix hack is because when sending numerics you have to parse the 
   * target differently, depending if its sent as <numnick> or :server.
   * if prefix_hack is 1, it was sent as :server, and numeric target
   * must be nick. Otherwise, prefix_hack is 0, and prefix was a proper
   * prefix and target is a numnick. (ircu2.10.08 sends :server)
   */
  int prefix_hack = 0;

  Sender = (char *) strtok(LineBuff, " ");
  if(Sender[0] == ':')
  {
     prefix_hack++;  /* See above comment */
     Sender++;
     if(!(ServerPtr = GetServer(Sender)))
     {
        if(!(UserPtr = FindNick(Sender, UsersN)))
           Debug(DBGINFO, "Sender %s not found", Sender);
     }
  }
  else 
  {  /* Find by numnicks */
     if(!((!Sender[1]) && (ServerPtr = GetServerByNumnick(Sender[0]))))
     {
       if(!(UserPtr = FindNumnick(Sender, UsersU)))
           Debug(DBGINFO, "Sender %s not found", Sender);
     }
  }
  Command = (char *) strtok(NULL, " ");
  rest = (char *) strtok(NULL, "\r\n\0");
/*** Check if maybe its a numeric ***/
  if(isdigit(Command[0]) && ServerPtr)  /* If first letter is a number and a server sent it.. */
  {
     int num = 0;
     if((num = atoi(Command)))   /* if num is non zero.. */
     {
         char* target;
         UserListType TmpUserPtr;
         target = (char *) strtok(rest, " ");
         rest = (char *) strtok(NULL, "\r\n");
         if(prefix_hack)
            TmpUserPtr = FindNick(target, UsersN);
         else
            TmpUserPtr = FindNumnick(target, UsersU);
         ProcessNumeric(ServerPtr,  num, TmpUserPtr, rest);
         return(0);
     }
  }
  for(i=0;server_commands[i].name;i++)
  {
      if(!strcasecmp(server_commands[i].name, Command))
      {
         if(UserPtr)
         {
             if(server_commands[i].usercmd == FALSE)
             {
               Debug(DBGINFO, "ERROR: User tried to %s", Command);
               return(0);
             }
         }
         else if(ServerPtr)
         {
             if(server_commands[i].servercmd == FALSE)
             {
                Debug(DBGINFO, "ERROR: Server tried to %s", Command);
                return(0);
             }
         }
         else
         {
             Debug(DBGWARNING, "Command called but source (%s) is neither user nor server", Sender);
             Debug(DBGWARNING, "This means someone broke protocol. Im bailing..");
             SockConnected = FALSE;
             /* This can be triggered by ircu when a P9 server is connected
              * but doesnt have proper configuration to map to a numeric.
              * the server will start sending stuff without its numnick.
              */
             return(0);
         }
         server_commands[i].function(ServerPtr, UserPtr, rest);
         return(0);
      }
  }
  Debug(DBGWARNING, "Please teach me how to parse ..%s %s", Command, rest);
  return(0);
}

int ProcessLine(char *LineBuff)
{
  Debug(DBGNOTICE, "<- %s", LineBuff);
  TotalLines++; /* for profileing type info  */
  GotPong++; /* Anything from the socket is good enough for me... */

  /* For P10 since it is in transition, there are 4 possible things
   * we can recieve from the irc server. 
   * :server.afternet.org <stuff..>  normal old p09 type msg
   * :nick <stuf..> normal old nick type msg
   * C <stuff..>  server numnick based command
   * CAA <stuff>   user numnick based command
   */
   if(LineBuff[0] == ':')
   {
     /* Parse old type prefixd commands */
     ProcessMessage(LineBuff);
   }
   else
   {
     /* Now here, it could be we are getting
      * a numnick prefix'd line, or a simple
      * line such as 
      * PASS :foo
      * To tell which we are expecting, everything
      * will be assumed to be PASS SERVER etc untill
      * we recieve the SERVER and then everything is
      * numnick.
      */
       if(REMOTE_AUTHENTICATED)
       { /* Must be a numnick */
          ProcessMessage(LineBuff);
       }
       else /* remote NOT yet authenticated */
       { /* Must be PASS SERVER etc. */
          ProcessPreAuthLine(LineBuff);
       }
   }
   return(0);
}

/*
 * Gets data from the socket.. doesnt care what it is. 
 * returns the number of bytes gotten.. the data is put in Buffer
 */
int ReadSocket(int TheSocket, char *Buffer)
{
   int c;
   int i;
   static char remainder[MAXRECVLEN+10] = "";
   static char LineString[MAXRECVLEN+10]; /* TODO(relnev) - large things should not be on the stack */
   int rsize;
   char *newline;
   char *doneline;
   
   /* Clear out the leftovers buffer, we got a new connection */
   if(TheSocket == -1)
   {
     memset(remainder, '\0', MAXRECVLEN);
     return(0);
   }
   memset(Buffer, '\0', MAXRECVLEN+1);
   rsize = 0;
   if(*remainder)
   {
      rsize = strlen(remainder);
      memcpy(Buffer, remainder, rsize);
   }
   c = recv(TheSocket, &Buffer[rsize], MAXRECVLEN-rsize, 0);
   if(c <= 0)
   {
       if (c == 0)
           Debug(DBGWARNING, "0 bytes from socket. Killing..");
       else
           Debug(DBGWARNING, "Got -1 from recv (%s). Killing..", strerror(errno));
           
       remainder[0] = '\0';
       SockConnected = FALSE;
       return(0);
   }
   newline = Buffer;
   c = c+rsize;
   Buffer[c] = '\0';
   for(i=0;i<c;i++)
   {
      if(Buffer[i] == '\n')
      {
         Buffer[i] = '\0';
         memcpy(LineString, newline, strlen(newline)+1);
         ProcessLine(LineString);
         doneline = newline;
         newline = &Buffer[i+1]; 
      }
      if(Buffer[i] == '\r')
         Buffer[i] = '\0';
   }
   if(*newline)
   {
      strcpy(remainder, newline);
   }
   else
   {
      memset(remainder, '\0', MAXRECVLEN);
   }
   return(1);
}

